-- some basic options
local vim = vim
vim.opt.tabstop = 4
vim.opt.shiftwidth = 4
vim.opt.softtabstop = 2
vim.opt.expandtab = true
vim.opt.termguicolors = true
vim.opt.autoindent = true
vim.opt.smartindent = true
vim.opt.smarttab = true
vim.opt.number = true
vim.o.hidden = true
vim.o.signcolumn = 'auto'
vim.o.breakindent = true
vim.o.smarttab = true
vim.o.errorbells = true
vim.o.clipboard = "unnamedplus"
vim.g.noswapfile = true
vim.cmd [[set mouse=a]] -- enable mouse
vim.wo.cursorline = true
vim.opt.hlsearch = false
vim.opt.wrap = true
vim.opt.showmatch = true

vim.api.nvim_create_user_command('W', 'write', {})

vim.opt.foldmethod = "expr"
vim.opt.foldexpr = "nvim_treesitter#foldexpr()"

-- hit zc to enable folding
vim.o.foldenable = false

vim.opt.guifont = "Iosevka Nerd Font:h12"

vim.g.neovide_refresh_rate = 140
vim.g.neomake_open_list = 2

function map(mode, shortcut, command)
    vim.api.nvim_set_keymap(mode, shortcut, command, { noremap = true, silent = true })
end

-- normal mode map
function nmap(shortcut, command)
    map('n', shortcut, command)
end

-- insert mode map
function imap(shortcut, command)
    map('i', shortcut, command)
end

-- <leader> is leader key; default is \
nmap("<leader>v", "<cmd>NvimTreeToggle<cr>")
nmap("<leader>ec", "<cmd>e ~/.config/nvim/init.lua<cr>")
nmap("<leader>sc", "<cmd>source ~/.config/nvim/init.lua<cr>")

-- buffer commands
nmap("<leader>b", "<cmd>BufferLineCyclePrev<cr>")
nmap("<leader>n", "<cmd>BufferLineCycleNext<cr>")
nmap("<leader>gb", "<cmd>BufferLinePick<cr>")
nmap("<leader>w", "<cmd>bdelete!<cr>")

-- toggle terminal in both modes
nmap("<leader>t", "<cmd>FloatermToggle<cr>")
map("t", "<leader>t", "<cmd>FloatermToggle<cr>")

-- lsp on_attach mappings

local augroup = vim.api.nvim_create_augroup("LspFormatting", {})
local on_attach = function(client, bufnr)
    -- Enable completion triggered by <c-x><c-o>
    vim.api.nvim_buf_set_option(bufnr, 'omnifunc', 'v:lua.vim.lsp.omnifunc')

    -- Mappings.
    -- See `:help vim.lsp.*` for documentation on any of the below functions
    local bufopts = { noremap = true, silent = true, buffer = bufnr }
    vim.keymap.set('n', 'gD', vim.lsp.buf.declaration, bufopts)
    vim.keymap.set('n', 'gd', vim.lsp.buf.definition, bufopts)
    vim.keymap.set('n', 'K', vim.lsp.buf.hover, bufopts)
    vim.keymap.set('n', 'gi', vim.lsp.buf.implementation, bufopts)
    vim.keymap.set('n', '<C-k>', vim.lsp.buf.signature_help, bufopts)
    vim.keymap.set('n', '<space>wa', vim.lsp.buf.add_workspace_folder, bufopts)
    vim.keymap.set('n', '<space>wr', vim.lsp.buf.remove_workspace_folder, bufopts)
    vim.keymap.set('n', '<space>wl', function()
        print(vim.inspect(vim.lsp.buf.list_workspace_folders()))
    end, bufopts)
    vim.keymap.set('n', '<space>D', vim.lsp.buf.type_definition, bufopts)
    vim.keymap.set('n', '<space>rn', vim.lsp.buf.rename, bufopts)
    vim.keymap.set('n', '<space>ca', vim.lsp.buf.code_action, bufopts)
    vim.keymap.set('n', 'gr', vim.lsp.buf.references, bufopts)
    vim.keymap.set('n', '<space>f', vim.lsp.buf.formatting, bufopts)

    require "coq".lsp_ensure_capabilities {}
    require 'illuminate'.on_attach(client)

    -- format on save
    if client.supports_method("textDocument/formatting") then

        -- get filetype
        local filetype = vim.api.nvim_buf_get_option(bufnr, "filetype")
        -- get null-ls supported sources
        local available = require "null-ls.sources".get_available(filetype, require("null-ls").methods.FORMATTING)

        local enable = false
        if #available > 0 then
            enable = client.name == "null-ls"
        else
            enable = not (client.name == "null-ls")
        end

        client.resolved_capabilities.document_formatting = enable
        client.resolved_capabilities.document_range_formatting = enable

        if client.resolved_capabilities.document_formatting then
            vim.api.nvim_create_autocmd("BufWritePre", {
                group = augroup,
                buffer = bufnr,
                callback = function()
                    vim.lsp.buf.formatting_sync()
                end
            })
        end
    end
end

-- lsp config setup
local lsp = require "lspconfig"
lsp.hls.setup { on_attach = on_attach } -- haskell language server
lsp.pyright.setup { on_attach = on_attach }
lsp.tsserver.setup { on_attach = on_attach }
lsp.rust_analyzer.setup { on_attach = on_attach }
lsp.sumneko_lua.setup { on_attach = on_attach,
    settings = {
        Lua = {
            diagnostics = {
                globals = { 'vim' }
            }
        }
    }
}

-- lsp diagnostic mapping
local opts = { noremap = true, silent = true }
vim.keymap.set('n', '<space>e', vim.diagnostic.open_float, opts)
vim.keymap.set('n', '[d', vim.diagnostic.goto_prev, opts)
vim.keymap.set('n', ']d', vim.diagnostic.goto_next, opts)
vim.keymap.set('n', '<space>q', vim.diagnostic.setloclist, opts)

vim.cmd("colorscheme carbonfox")

-- packer boilerplate
local fn = vim.fn
local install_path = fn.stdpath('data') .. '/site/pack/packer/start/packer.nvim'
if fn.empty(fn.glob(install_path)) > 0 then
    packer_bootstrap = fn.system({ 'git', 'clone', '--depth', '1', 'https://github.com/wbthomason/packer.nvim',
        install_path })
end

return require('packer').startup(function(use)
    use 'wbthomason/packer.nvim'
    use 'nvim-lua/plenary.nvim'
    use 'neovim/nvim-lspconfig'
    use {
        'ms-jpq/coq_nvim',
        branch = 'coq',
        event = "VimEnter",
        config = 'vim.cmd[[COQnow]]'
    }
    use { 'ms-jpq/coq.artifacts', branch = 'artifacts' }
    use { 'ms-jpq/coq.thirdparty', branch = '3p' }
    use {
        'nvim-treesitter/nvim-treesitter',
        run = ':TSUpdate',
        config = function() require('nvim-treesitter.configs').setup {
                highlight = {
                    enable = true
                },
            }
        end
    }
    use { 'kyazdani42/nvim-tree.lua',
        requires = {
            'kyazdani42/nvim-web-devicons'
        },
        config = function() require('nvim-tree').setup {} end
    }
    use {
        'jose-elias-alvarez/null-ls.nvim',
        config = function() require('null-ls').setup({
                sources = {
                    require('null-ls').builtins.formatting.eslint_d,
                    require('null-ls').builtins.diagnostics.eslint_d,
                    require("null-ls").builtins.code_actions.eslint_d,
                    require("null-ls").builtins.formatting.black
                }
            })
        end,
        requires = { "nvim-lua/plenary.nvim" },
    }
    use {
        'folke/trouble.nvim',
        config = function() require("trouble").setup {}
        end
    }
    use {
        'akinsho/bufferline.nvim',
        tag = "v2.*",
        requires = 'kyazdani42/nvim-web-devicons',
        config = function() require("bufferline").setup {
                options = {
                    diagnostics = "nvim_lsp",
                }
            }
        end
    }
    use { "EdenEast/nightfox.nvim",
        config = function() require("nightfox").setup {}
        end
    }
    use {
        "nvim-lualine/lualine.nvim",
        event = "VimEnter",
        config = function()
            require("lualine").setup {
                options = { theme = 'auto', globalstatus = true },
                sections = { lualine_x = { 'filetype' } }
            }
        end,
        requires = { 'kyazdani42/nvim-web-devicons', opt = true }
    }
    use {
        "startup-nvim/startup.nvim",
        requires = { "nvim-telescope/telescope.nvim", "nvim-lua/plenary.nvim" },
        config = function()
            require "startup".setup()
        end
    }
    use 'RRethy/vim-illuminate'
    use {
        'nvim-telescope/telescope.nvim',
        requires = { 'nvim-lua/plenary.nvim' },
        config = function()
            require("telescope").setup {}
        end
    }
    use {
        'j-hui/fidget.nvim',
        config = function()
            require('fidget').setup {}
        end
    }
    use 'voldikss/vim-floaterm'
    use {
        'kkoomen/vim-doge',
        run = ':call doge#install()'
    }
    use { 'norcalli/nvim-colorizer.lua',
        config = function() require('colorizer').setup {
                'css';
                'javascript';
            }
        end,
    }
    use { 'lewis6991/gitsigns.nvim',
        config = function()
            require('gitsigns').setup()
        end
    }
    if packer_bootstrap then
        require('packer').sync()
    end
end)
